/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

/**
 * A backpointer to be used in CKY parsing. A backpointer specifies a recipe for
 * how to build a tree. It consists of a rule, which defines a tree of depth 1,
 * and up to two references to other backpointers, which recursively define the
 * immediate subtrees.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class Backpointer {

	/**
	 * The rule associated with this backpointer.
	 */
	private final Rule rule;
	/**
	 * The backpointer for the left subtree.
	 */
	private final Backpointer pointerL;
	/**
	 * The backpointer for the right subtree.
	 */
	private final Backpointer pointerR;

	/**
	 * Create a new backpointer based on a preterminal rule.
	 *
	 * @param rule a preterminal rule
	 */
	public Backpointer(PreterminalRule rule) {
		this.rule = rule;
		this.pointerL = null;
		this.pointerR = null;
	}

	/**
	 * Create a new backpointer based on a unary rule.
	 *
	 * @param rule a unary rule
	 * @param pointer a backpointer to the subtree
	 */
	public Backpointer(UnaryRule rule, Backpointer pointer) {
		assert pointer != null;
		this.rule = rule;
		this.pointerL = pointer;
		this.pointerR = null;
	}

	/**
	 * Create a new backpointer based on a binary rule.
	 *
	 * @param rule a binary rule
	 * @param pointerL a backpointer for the left subtree
	 * @param pointerR a backpointer for the right subtree
	 */
	public Backpointer(BinaryRule rule, Backpointer pointerL, Backpointer pointerR) {
		assert pointerL != null;
		assert pointerR != null;
		this.rule = rule;
		this.pointerL = pointerL;
		this.pointerR = pointerR;
	}

	/**
	 * Returns the rule associated with this backpointer.
	 *
	 * @return the rule associated with this backpointer
	 */
	public Rule getRule() {
		return rule;
	}

	/**
	 * Returns the backpointer for the left subtree.
	 *
	 * @return the backpointer for the left subtree, or <code>null</code> in
	 * case no such backpointer is specified
	 */
	public Backpointer getPointerL() {
		return pointerL;
	}

	/**
	 * Returns the backpointer for the right subtree.
	 *
	 * @return the backpointer for the right subtree, or <code>null</code> in
	 * case no such backpointer is specified
	 */
	public Backpointer getPointerR() {
		return pointerR;
	}

	/**
	 * Return the tree specified by this backpointer. The rules used in the
	 * backpointer and its substructures will be interpreted with respect to the
	 * specified grammar.
	 *
	 * @param grammar a grammar
	 * @return the tree specified by this backpointer
	 */
	public Tree extract(Grammar grammar) {
		Tree tree = new Tree(grammar.getCategory(rule.getLHS()));
		if (rule.getClass() == PreterminalRule.class) {
			PreterminalRule preterminalRule = (PreterminalRule) rule;
			String label = grammar.getWord(preterminalRule.getRHS());
			tree.addAsRightmostSubtree(new Tree(label));
		}
		if (rule.getClass() == UnaryRule.class) {
			tree.addAsRightmostSubtree(pointerL.extract(grammar));
		}
		if (rule.getClass() == BinaryRule.class) {
			BinaryRule binaryRule = (BinaryRule) rule;
			tree.addAsRightmostSubtree(pointerL.extract(grammar));
			tree.addAsRightmostSubtree(pointerR.extract(grammar));
		}
		return tree;
	}
}
