/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

/**
 * A chart that holds backpointers.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class BackpointerChart {

	/**
	 * The low-level chart, represented as a (ragged) array.
	 */
	private final Backpointer[][][] chart;

	/**
	 * Construct a new chart.
	 *
	 * @param nCategories the number of categories that the chart should hold
	 * @param nTokens the number of tokens that the chart should hold
	 */
	public BackpointerChart(int nCategories, int nTokens) {
		this.chart = new Backpointer[nTokens + 1][][];
		for (int i = nTokens; i >= 0; i--) {
			chart[i] = new Backpointer[nTokens + 1 - i][nCategories];
		}
	}

	/**
	 * Returns the backpointer for the specified span in this chart.
	 *
	 * @param cat the syntactic category of the span
	 * @param min the leftmost position of the span
	 * @param max the rightmost position of the span
	 * @return the backpointer for the specified span, or <code>null</code> in
	 * case this chart does not contain a backpointer for the specified span
	 */
	public Backpointer get(int cat, int min, int max) {
		return chart[min][max - min][cat];
	}

	/**
	 * Replaces the backpointer for the specified span in this chart with the
	 * specified value.
	 *
	 * @param cat the syntactic category of the span
	 * @param min the leftmost position of the span
	 * @param max the rightmost position of the span
	 * @param bp the new backpointer for the specified span
	 */
	public void set(int cat, int min, int max, Backpointer bp) {
		chart[min][max - min][cat] = bp;
	}
}
