/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

/**
 * Default implementation of the
 * <code>TreeHandler</code> interface. This implementation provides
 * functionality to build simple pipes of tree handlers, where a given handler
 * can pass on notifications to its successor in the pipe.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class DefaultTreeHandler implements TreeHandler {

	/**
	 * The successor of this handler in the handler chain.
	 */
	private TreeHandler successor;

	/**
	 * Constructs a new handler with the specified successor.
	 *
	 * @param successor the successor of this handler
	 */
	public DefaultTreeHandler(TreeHandler successor) {
		this.successor = successor;
	}

	/**
	 * Constructs a new handler with no successor.
	 */
	public DefaultTreeHandler() {
		successor = NullTreeHandler.getInstance();
	}

	/**
	 * Returns the successor of this handler.
	 *
	 * @return the successor of this handler
	 */
	public TreeHandler getSuccessor() {
		return successor;
	}

	/**
	 * Replaces the successor of this handler with the specified handler.
	 *
	 * @param handler the new successor of this handler
	 */
	public void setSuccessor(TreeHandler handler) {
		assert handler != null;
		successor = handler;
	}

	/**
	 * Notifies the successor of this handler about the start of the stream.
	 */
	public void init() {
		successor.init();
	}

	/**
	 * Notifies the successor of this handler about the specified tree.
	 *
	 * @param tree the tree about which the successor of this handler is to be
	 * notified
	 */
	public void next(Tree tree) {
		successor.next(tree);
	}

	/**
	 * Notifies the successor of this handler about the end of the stream.
	 */
	public void exit() {
		successor.exit();
	}
}
