/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * A probabilistic context-free grammar.
 *
 * <p>When a rule is added to the grammar, the rule itself as well as all
 * terminal and nonterminal symbols occurring in the rule are assigned a
 * <em>code</em>, a unique, non-negative integer. Two rules or symbols will
 * receive the same code if and only if they are equal with respect to the
 * method {@code equals}.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class Grammar {

	/**
	 * The root category (start symbol) of this grammar.
	 */
	private static final String ROOT = "ROOT";
	/**
	 * The unknown word.
	 */
	private static final String UNKNOWN = "UNK";
	/**
	 * The category index used by this grammar.
	 */
	private final Index<String> categoryIndex;
	/**
	 * The word index used by this grammar.
	 */
	private final Index<String> wordIndex;
	/**
	 * The preterminal rules used by this grammar.
	 */
	private final Index<PreterminalRule> preterminalRules;
	/**
	 * The preterminal rules by their left-hand side.
	 */
	private final ArrayList<List<PreterminalRule>> preterminalRulesByLHS;
	/**
	 * The preterminal rules by their right-hand side.
	 */
	private final ArrayList<List<PreterminalRule>> preterminalRulesByRHS;
	/**
	 * The unary rules used by this grammar.
	 */
	private final Index<UnaryRule> unaryRules;
	/**
	 * The unary rules by their left-hand side.
	 */
	private final ArrayList<List<UnaryRule>> unaryRulesByLHS;
	/**
	 * The unary rules by their right-hand side.
	 */
	private final ArrayList<List<UnaryRule>> unaryRulesByRHS;
	/**
	 * The binary rules used by this grammar.
	 */
	private final Index<BinaryRule> binaryRules;
	/**
	 * The binary rules by their left-hand side.
	 */
	private final ArrayList<List<BinaryRule>> binaryRulesByLHS;
	/**
	 * The binary rules by the first nonterminal on their right-hand side.
	 */
	private final ArrayList<List<BinaryRule>> binaryRulesByRHS1;
	/**
	 * The binary rules by the second nonterminal on their right-hand side.
	 */
	private final ArrayList<List<BinaryRule>> binaryRulesByRHS2;
	/**
	 * Whether the grammar is sealed (not open to new rules).
	 */
	private boolean isSealed;

	/**
	 * Constructs a new grammar.
	 */
	public Grammar() {
		this.categoryIndex = new Index<String>();
		categoryIndex.add(ROOT);

		this.wordIndex = new Index<String>();
		wordIndex.add(UNKNOWN);

		this.preterminalRules = new Index<PreterminalRule>();
		this.preterminalRulesByLHS = new ArrayList<List<PreterminalRule>>();
		this.preterminalRulesByRHS = new ArrayList<List<PreterminalRule>>();

		this.unaryRules = new Index<UnaryRule>();
		this.unaryRulesByLHS = new ArrayList<List<UnaryRule>>();
		this.unaryRulesByRHS = new ArrayList<List<UnaryRule>>();

		this.binaryRules = new Index<BinaryRule>();
		this.binaryRulesByLHS = new ArrayList<List<BinaryRule>>();
		this.binaryRulesByRHS1 = new ArrayList<List<BinaryRule>>();
		this.binaryRulesByRHS2 = new ArrayList<List<BinaryRule>>();

		this.isSealed = false;
	}

	/**
	 * Returns the number of categories of this grammar.
	 *
	 * @return the number of categories of this grammar
	 */
	public int getNCategories() {
		return categoryIndex.getNEntries();
	}

	/**
	 * Returns the number of words of this grammar.
	 *
	 * @return the number of words of this grammar
	 */
	public int getNWords() {
		return wordIndex.getNEntries();
	}

	/**
	 * Returns the code of the specified syntactic category.
	 *
	 * @param category the syntactic category whose code to return
	 * @return the code of the specified category, or <code>-1</code> in case
	 * the specified category is not a category of this grammar
	 */
	public int getCodeCategory(String category) {
		return categoryIndex.getIndex(category);
	}

	/**
	 * Returns the code of the root category of this grammar.
	 *
	 * @return the code of the root category (start symbol) category of this
	 * grammar
	 */
	public int getRootCode() {
		return getCodeCategory(ROOT);
	}

	/**
	 * Returns the category of this grammar with the specified code.
	 *
	 * @param code the code whose corresponding category to return
	 * @return the category of this grammar with the specified code, * *
	 * or <code>null</code> in case this grammar does not have a category with
	 * the specified code
	 */
	public String getCategory(int code) {
		return categoryIndex.getEntry(code);
	}

	/**
	 * Returns the code of the specified word.
	 *
	 * @param word the word whose code to return
	 * @return the code of the specified word, or <code>-1</code> in case the
	 * specified word is not a word of this grammar
	 */
	public int getCodeWord(String word) {
		return wordIndex.getIndex(word);
	}

	/**
	 * Returns the code of the unknown word.
	 *
	 * @return the code of the unknown word
	 */
	public int getUnknownCode() {
		return getCodeWord(UNKNOWN);
	}

	/**
	 * Returns the word with the specified code.
	 *
	 * @param code the code whose corresponding word to return
	 * @return the word of this grammar with the specified code, *
	 * or <code>null</code> in case this grammar does not have a word with the
	 * specified code
	 */
	public String getWord(int code) {
		return wordIndex.getEntry(code);
	}

	/**
	 * Adds the specified preterminal rule to this grammar.
	 *
	 * @param lhs the left-hand side category (POS tag) of the rule
	 * @param rhs the right-hand side (word) of the rule
	 * @return the preterminal rule
	 */
	public PreterminalRule addPreterminalRule(String lhs, String rhs) {
		if (isSealed) {
			throw new RuntimeException("This grammar does not allow the addition of new rules.");
		} else {
			PreterminalRule rule = new PreterminalRule(categoryIndex.add(lhs), wordIndex.add(rhs));
			return preterminalRules.internalize(rule);
		}
	}

	/**
	 * Adds the specified unary rule to this grammar.
	 *
	 * @param lhs the left-hand side category of the rule
	 * @param rhs the right-hand side category of the rule
	 * @return the unary rule
	 */
	public UnaryRule addUnaryRule(String lhs, String rhs) {
		if (isSealed) {
			throw new RuntimeException("This grammar does not allow the addition of new rules.");
		} else {
			UnaryRule rule = new UnaryRule(categoryIndex.add(lhs), categoryIndex.add(rhs));
			return unaryRules.internalize(rule);
		}
	}

	/**
	 * Adds the specified binary rule to this grammar.
	 *
	 * @param lhs the left-hand side category of the rule
	 * @param rhs1 the first category on the right-hand side of the rule
	 * @param rhs2 the second category on the right-hand side of the rule
	 * @return the binary rule
	 */
	public BinaryRule addBinaryRule(String lhs, String rhs1, String rhs2) {
		if (isSealed) {
			throw new RuntimeException("This grammar does not allow the addition of new rules.");
		} else {
			BinaryRule rule = new BinaryRule(categoryIndex.add(lhs), categoryIndex.add(rhs1), categoryIndex.add(rhs2));
			return binaryRules.internalize(rule);
		}
	}

	/**
	 * Seals the grammar. This closes the grammar for the addition of new rules.
	 */
	public void seal() {
		rebuildPreterminalRulesIndexes();
		rebuildUnaryRulesIndexes();
		rebuildBinaryRulesIndexes();
		isSealed = true;
	}

	/**
	 * Returns the number of rules of this grammar.
	 *
	 * @return the number of rules of this grammar.
	 */
	public int getNRules() {
		return getNPreterminalRules() + getNUnaryRules() + getNBinaryRules();
	}

	/**
	 * Returns the preterminal rules of this grammar.
	 *
	 * @return the preterminal rules of this grammar
	 */
	public List<PreterminalRule> getPreterminalRules() {
		return preterminalRules.getEntries();
	}

	/**
	 * Returns the number of preterminal rules of this grammar.
	 *
	 * @return the number of preterminal rules of this grammar
	 */
	public int getNPreterminalRules() {
		return preterminalRules.getNEntries();
	}

	/**
	 * Returns the preterminal rules of this grammar with the specified
	 * left-hand side.
	 *
	 * @param lhs the code of the left-hand side (category)
	 * @return the preterminal rules of this grammar with the specified
	 * left-hand side
	 */
	public List<PreterminalRule> getPreterminalRulesWithLHS(int lhs) {
		return preterminalRulesByLHS.get(lhs);
	}

	/**
	 * Returns the preterminal rules of this grammar with the specified
	 * right-hand side.
	 *
	 * @param rhs the code of the right-hand side (word)
	 * @return the preterminal rules of this grammar with the specified
	 * right-hand side
	 */
	public List<PreterminalRule> getPreterminalRulesWithRHS(int rhs) {
		return preterminalRulesByRHS.get(rhs);
	}

	/**
	 * Rebuild the indexes for the preterminal rules.
	 */
	private void rebuildPreterminalRulesIndexes() {
		preterminalRulesByLHS.clear();
		preterminalRulesByRHS.clear();
		for (int i = 0; i < getNCategories(); i++) {
			preterminalRulesByLHS.add(new LinkedList<PreterminalRule>());
		}
		for (int i = 0; i < getNWords(); i++) {
			preterminalRulesByRHS.add(new LinkedList<PreterminalRule>());
		}
		preterminalRulesByLHS.trimToSize();
		preterminalRulesByRHS.trimToSize();

		for (PreterminalRule rule : preterminalRules.getEntries()) {
			preterminalRulesByLHS.get(rule.getLHS()).add(rule);
			preterminalRulesByRHS.get(rule.getRHS()).add(rule);
		}
	}

	/**
	 * Returns the unary rules of this grammar.
	 *
	 * @return the unary rules of this grammar
	 */
	public List<UnaryRule> getUnaryRules() {
		return unaryRules.getEntries();
	}

	/**
	 * Returns the number of unary rules of this grammar.
	 *
	 * @return the number of unary rules of this grammar
	 */
	public int getNUnaryRules() {
		return unaryRules.getNEntries();
	}

	/**
	 * Returns the unary rules of this grammar with the specified left-hand
	 * side.
	 *
	 * @param lhs the code of the left-hand side
	 * @return the unary rules of this grammar with the specified left-hand side
	 */
	public List<UnaryRule> getUnaryRulesWithLHS(int lhs) {
		return unaryRulesByLHS.get(lhs);
	}

	/**
	 * Returns the unary rules of this grammar with the specified right-hand
	 * side.
	 *
	 * @param rhs the code of the right-hand side
	 * @return the unary rules of this grammar with the specified right-hand
	 * side
	 */
	public List<UnaryRule> getUnaryRulesWithRHS(int rhs) {
		return unaryRulesByRHS.get(rhs);
	}

	/**
	 * Rebuild the indexes for the unary rules.
	 */
	private void rebuildUnaryRulesIndexes() {
		unaryRulesByLHS.clear();
		unaryRulesByRHS.clear();
		for (int i = 0; i < getNCategories(); i++) {
			unaryRulesByLHS.add(new LinkedList<UnaryRule>());
			unaryRulesByRHS.add(new LinkedList<UnaryRule>());
		}
		unaryRulesByLHS.trimToSize();
		unaryRulesByRHS.trimToSize();

		for (UnaryRule rule : unaryRules.getEntries()) {
			unaryRulesByLHS.get(rule.getLHS()).add(rule);
			unaryRulesByRHS.get(rule.getRHS()).add(rule);
		}
	}

	/**
	 * Returns the binary rules of this grammar.
	 *
	 * @return the binary rules of this grammar
	 */
	public List<BinaryRule> getBinaryRules() {
		return binaryRules.getEntries();
	}

	/**
	 * Returns the number of binary rules of this grammar.
	 *
	 * @return the number of binary rules of this grammar
	 */
	public int getNBinaryRules() {
		return binaryRules.getNEntries();
	}

	/**
	 * Returns the binary rules of this grammar with the specified left-hand
	 * side.
	 *
	 * @param lhs the left-hand side
	 * @return the binary rules of this grammar with the specified left-hand
	 * side
	 */
	public List<BinaryRule> getBinaryRulesWithLHS(int lhs) {
		return binaryRulesByLHS.get(lhs);
	}

	/**
	 * Returns the binary rules of this grammar with the specified right-hand
	 * side.
	 *
	 * @param rhs1 the first category on the right-hand side
	 * @return the binary rules of this grammar that have the specified category
	 * as the first category on their right-hand side
	 */
	public List<BinaryRule> getBinaryRulesWithRHS1(int rhs1) {
		return binaryRulesByRHS1.get(rhs1);
	}

	/**
	 * Returns the binary rules of this grammar with the specified right-hand
	 * side.
	 *
	 * @param rhs2 the second category on the right-hand side
	 * @return the binary rules of this grammar that have the specified category
	 * as the second category on their right-hand side
	 */
	public List<BinaryRule> getBinaryRulesWithRHS2(int rhs2) {
		return binaryRulesByRHS2.get(rhs2);
	}

	/**
	 * Rebuild the indexes for the binary rules.
	 */
	private void rebuildBinaryRulesIndexes() {
		binaryRulesByLHS.clear();
		binaryRulesByRHS1.clear();
		binaryRulesByRHS2.clear();
		for (int i = 0; i < getNCategories(); i++) {
			binaryRulesByLHS.add(new LinkedList<BinaryRule>());
			binaryRulesByRHS1.add(new LinkedList<BinaryRule>());
			binaryRulesByRHS2.add(new LinkedList<BinaryRule>());
		}
		binaryRulesByLHS.trimToSize();
		binaryRulesByRHS1.trimToSize();
		binaryRulesByRHS2.trimToSize();

		for (BinaryRule rule : binaryRules.getEntries()) {
			binaryRulesByLHS.get(rule.getLHS()).add(rule);
			binaryRulesByRHS1.get(rule.getRHS1()).add(rule);
			binaryRulesByRHS2.get(rule.getRHS2()).add(rule);
		}
	}
}
