/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

/**
 * Extract a PCFG from a stream of trees.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class GrammarExtractor extends DefaultTreeHandler {

	/**
	 * The unknown word.
	 */
	private static final String UNKNOWN = "UNK";
	/**
	 * The grammar constructed.
	 */
	private Grammar grammar;

	/**
	 * Constructs a new handler.
	 */
	public GrammarExtractor() {
	}

	/**
	 * Returns the extracted grammar.
	 *
	 * @return the grammar extracted by this handler
	 */
	public Grammar getGrammar() {
		return grammar;
	}

	@Override
	public void init() {
		this.grammar = new Grammar();

		// Hack for unknown words. Figures taken from
		// http://stp.ling.uu.se/~nivre/master/statmet4.html
		grammar.addPreterminalRule("CD", UNKNOWN).setScore(696);
		grammar.addPreterminalRule("JJ", UNKNOWN).setScore(1184);
		grammar.addPreterminalRule("NN", UNKNOWN).setScore(1308);
		grammar.addPreterminalRule("NNP", UNKNOWN).setScore(1799);
		grammar.addPreterminalRule("NNS", UNKNOWN).setScore(724);

		super.init();
	}

	@Override
	public void next(Tree tree) {
		tree.traverseLeftToRight(new Handler());
		super.next(tree);
	}

	@Override
	public void exit() {
		// Do not allow the addition of further rules.
		grammar.seal();

		// Accumulate rule counts of rules by lhs.
		double[] ruleCountsByLHS = new double[grammar.getNCategories()];
		for (PreterminalRule rule : grammar.getPreterminalRules()) {
			ruleCountsByLHS[rule.getLHS()] += rule.getScore();
		}
		for (UnaryRule rule : grammar.getUnaryRules()) {
			ruleCountsByLHS[rule.getLHS()] += rule.getScore();
		}
		for (BinaryRule rule : grammar.getBinaryRules()) {
			ruleCountsByLHS[rule.getLHS()] += rule.getScore();
		}

		// Compute relative frequencies.
		for (PreterminalRule rule : grammar.getPreterminalRules()) {
			rule.setScore(rule.getScore() / ruleCountsByLHS[rule.getLHS()]);
		}
		for (UnaryRule rule : grammar.getUnaryRules()) {
			rule.setScore(rule.getScore() / ruleCountsByLHS[rule.getLHS()]);
		}
		for (BinaryRule rule : grammar.getBinaryRules()) {
			rule.setScore(rule.getScore() / ruleCountsByLHS[rule.getLHS()]);
		}

		super.exit();

		System.err.format("Extracted a grammar with %d categories, %d words, and %d rules.%n", grammar.getNCategories(), grammar.getNWords(), grammar.getNRules());
	}

	/**
	 * Handler used to collect the rules from a phrase-structure tree.
	 */
	private class Handler implements TraversalHandler {

		public void enter(Tree tree) {
			if (!tree.isTerminal()) {
				String[] labels = new String[tree.getNSubtrees() + 1];
				labels[0] = tree.getLabel();
				int i = 1;
				for (Tree subtree : tree.getSubtrees()) {
					labels[i] = subtree.getLabel();
					assert labels[i] != null;
					assert labels[i].length() > 0;
					i++;
				}
				if (isPreterminal(tree)) {
					PreterminalRule rule = grammar.addPreterminalRule(labels[0], labels[1]);
					updateCount(rule);
				} else {
					switch (tree.getNSubtrees()) {
						case 1: {
							UnaryRule rule = grammar.addUnaryRule(labels[0], labels[1]);
							updateCount(rule);
						}
						break;
						case 2: {
							BinaryRule rule = grammar.addBinaryRule(labels[0], labels[1], labels[2]);
							updateCount(rule);
						}
						break;
						default:
							throw new RuntimeException("Can only process unary or binary branching trees.");
					}
				}
			}
		}

		public void leave(Tree tree) {
		}

		private boolean isPreterminal(Tree tree) {
			return tree.getNSubtrees() == 1 && tree.getSubtrees().getFirst().isTerminal();
		}
	}

	/**
	 * Updates the count of the specified rule. This misuses the rule's score to
	 * store the count.
	 *
	 * @param rule the rule whose count to update
	 */
	private static void updateCount(Rule rule) {
		double count = rule.getScore();
		if (count < 0) {
			rule.setScore(1);
		} else {
			rule.setScore(count + 1);
		}
	}
}
