/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

import java.io.*;
import java.util.HashMap;
import java.util.Map;

/**
 * Command-line interface for the parser.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class Main {

	/**
	 * Default options.
	 */
	private static final Map<String, String> OPTIONS_MAP = new HashMap<String, String>();

	/**
	 * Private constructor to prevent instantiation.
	 */
	private Main() {
	}

	/**
	 * Main method.
	 *
	 * @param args the command-line arguments
	 * @throws FileNotFoundException if some file is not found
	 * @throws IOException if some error occurs during reading or writing
	 */
	public static void main(String[] args) throws FileNotFoundException, IOException {
		if (args.length == 0) {
			usage();
			System.exit(1);
		}
		if (args[0].equals("extract")) {
			extract(args);
			System.exit(0);
		}
		if (args[0].equals("help")) {
			if (args.length == 1) {
				usage();
				System.exit(1);
			}
			if (args[1].equals("extract")) {
				extractHelp();
				System.exit(0);
			}
			if (args[1].equals("parse")) {
				parseHelp();
				System.exit(0);
			}
			if (args[1].equals("prepare")) {
				prepareHelp();
				System.exit(0);
			}
		}
		if (args[0].equals("parse")) {
			parse(args);
			System.exit(0);
		}
		if (args[0].equals("prepare")) {
			prepare(args);
			System.exit(0);
		}
		usage();
		System.exit(1);
	}

	private static void usage() {
		System.err.println("Usage: up (extract|help|parse|prepare) [options]");
		System.err.println();
		System.err.println("Type 'up help (extract|parse|prepare)' for more specific usage.");
	}

	private static void extract(String[] args) throws FileNotFoundException, IOException {
		Map<String, String> options = parseCommandLine(args);

		String rootFileName = getValueOrUseDefault(options, "--root", "treebank-3/parsed/mrg/wsj/");
		String grammarFileName = getValueOrUseDefault(options, "--grammar", "grammar.txt");

		// training data: usually sections 02-21; here only section 02
		int min = Integer.parseInt(getValueOrUseDefault(options, "--min", "0200"));
		int max = Integer.parseInt(getValueOrUseDefault(options, "--max", "0299"));

		StandardPreprocessor preprocessor = new StandardPreprocessor();
		TransformerHandler binarizer = new TransformerHandler(TreeTransformers.getBinarizer());
		GrammarExtractor grammarExtractor = new GrammarExtractor();
		TreeCounter counter = new TreeCounter();

		preprocessor.setSuccessor(binarizer);
		binarizer.setSuccessor(grammarExtractor);
		grammarExtractor.setSuccessor(counter);

		PennTreebankReader reader = new PennTreebankReader(preprocessor);
		reader.read(new File(rootFileName), min, max);

		Grammar grammar = grammarExtractor.getGrammar();

		PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(grammarFileName)));
		for (int lhs = 0; lhs < grammar.getNCategories(); lhs++) {
			for (UnaryRule rule : grammar.getUnaryRulesWithLHS(lhs)) {
				writer.print(grammar.getCategory(rule.getLHS()));
				writer.print(" -> ");
				writer.print(grammar.getCategory(rule.getRHS()));
				writer.print(" ");
				writer.print(rule.getScore());
				writer.println();
			}
			for (BinaryRule rule : grammar.getBinaryRulesWithLHS(lhs)) {
				writer.print(grammar.getCategory(rule.getLHS()));
				writer.print(" -> ");
				writer.print(grammar.getCategory(rule.getRHS1()));
				writer.print(" ");
				writer.print(grammar.getCategory(rule.getRHS2()));
				writer.print(" ");
				writer.print(rule.getScore());
				writer.println();
			}
		}
		for (int rhs = 0; rhs < grammar.getNWords(); rhs++) {
			for (PreterminalRule rule : grammar.getPreterminalRulesWithRHS(rhs)) {
				writer.print(grammar.getCategory(rule.getLHS()));
				writer.print(" => ");
				writer.print(grammar.getWord(rule.getRHS()));
				writer.print(" ");
				writer.print(rule.getScore());
				writer.println();
			}
		}
		writer.close();
	}

	private static void extractHelp() {
		System.err.println("extract: Extract a grammar from a set of WSJ files.");
		System.err.println();
		System.err.println("Command-specific options:");
		System.err.println("  --grammar  file in which to store the grammar");
		System.err.println("  --min      minimal file number to extract from (default: 0200)");
		System.err.println("  --max      maximal file number to extract from (default: 0299)");
		System.err.println("  --root     root of the directory in which to look for WSJ files");
	}

	private static void parse(String[] args) throws FileNotFoundException, IOException {
		Map<String, String> options = parseCommandLine(args);

		String grammarFileName = getValueOrUseDefault(options, "--grammar", "grammar.txt");
		String inFileName = getValueOrUseDefault(options, "--infile", "validate.mrg");
		String outFileName = options.get("--outfile");

		Grammar grammar = GrammarReader.read(grammarFileName);

		ParserHandler parserHandler = new ParserHandler(grammar);
		PennTreebankReader reader = new PennTreebankReader(parserHandler);

		if (outFileName != null) {
			TreePrinter treePrinter = new TreePrinter(new File(outFileName));
			parserHandler.setSuccessor(treePrinter);
		}

		reader.read(new File(inFileName));
	}

	private static void parseHelp() {
		System.err.println("parse: Parse a tree file using a grammar.");
		System.err.println();
		System.err.println("Command-specific options:");
		System.err.println("  --grammar  file from which to read the grammar");
		System.err.println("  --infile   file containing the validation data");
		System.err.println("  --outfile  file to which the parse trees will be written");
	}

	private static void prepare(String[] args) throws FileNotFoundException, IOException {
		Map<String, String> options = parseCommandLine(args);

		String rootFileName = getValueOrUseDefault(options, "--root", "treebank-3/parsed/mrg/wsj/");
		String outFileName = getValueOrUseDefault(options, "--outfile", "validate.mrg");

		// development data: section 22
		int min = Integer.parseInt(getValueOrUseDefault(options, "--min", "2200"));
		int max = Integer.parseInt(getValueOrUseDefault(options, "--max", "2201"));

		StandardPreprocessor preprocessor = new StandardPreprocessor();
		TreePrinter treePrinter = new TreePrinter(new File(outFileName));

		preprocessor.setSuccessor(treePrinter);

		PennTreebankReader reader = new PennTreebankReader(preprocessor);
		reader.read(new File(rootFileName), min, max);
	}

	private static void prepareHelp() {
		System.err.println("prepare: Prepare a tree file from a set of WSJ files.");
		System.err.println();
		System.err.println("Command-specific options:");
		System.err.println("  --min      minimal file number to extract from (default: 2200)");
		System.err.println("  --max      maximal file number to extract from (default: 2209)");
		System.err.println("  --outfile  file used to store the trees");
		System.err.println("  --root     root of the directory in which to look for WSJ files");
	}

	private static Map<String, String> parseCommandLine(String[] args) {
		Map<String, String> map = new HashMap<String, String>(OPTIONS_MAP);
		for (int i = 0; i < args.length; i++) {
			if (args[i].startsWith("--")) {
				if (i + 1 < args.length && !args[i + 1].startsWith("--")) {
					map.put(args[i], args[i + 1]);
					i++;
				} else {
					map.put(args[i], null);
				}
			}
		}
		return map;
	}

	private static <K, V> V getValueOrUseDefault(Map<K, V> map, K key, V defaultValue) {
		if (map.containsKey(key)) {
			return map.get(key);
		} else {
			return defaultValue;
		}
	}

	private static class StandardPreprocessor extends DefaultTreeHandler {

		public StandardPreprocessor() {
		}

		@Override
		public void next(Tree tree) {
			Tree tree1 = TreeTransformers.getFunctionAnnotationStripper().transform(tree);
			Tree tree2 = TreeTransformers.getEmptySubtreeStripper().transform(tree1);
			Tree tree3 = TreeTransformers.getUnaryChainsSplicer().transform(tree2);
//			Tree tree4 = TreeTransformers.getParentAnnotator().transform(tree3);
			super.next(tree3);
		}
	}
}
