/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Parse the yields of a stream of trees.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class ParserHandler extends DefaultTreeHandler {

	/**
	 * String used for unknown words.
	 */
	private static final String UNKNOWN = "UNK";
	/**
	 * The grammar to be used for parsing.
	 */
	private final Grammar grammar;
	/**
	 * The parser used by this handler.
	 */
	private final Parser parser;
	/**
	 * The fallback parser. This parser is called whenever the primary parser
	 * cannot parse a tag sequence.
	 */
	private final Parser fallbackParser;
	/**
	 * A scorer used to evaluate the performance of the parser.
	 */
	private final BracketScorer bracketScorer;
	/**
	 * The number of sentences parsed.
	 */
	private int nSentences;
	/**
	 * The total time used for parsing.
	 */
	private long totalTime;

	/**
	 * Constructs a new handler based on the specified grammar.
	 *
	 * @param grammar the grammar to be used for parsing
	 */
	public ParserHandler(Grammar grammar) {
		super();
		this.grammar = grammar;
//		this.parser = new SimpleParser(grammar);
        this.parser = new MyParser(grammar);
		this.fallbackParser = new StupidParser(grammar);
		this.bracketScorer = new BracketScorer();
	}

	@Override
	public void next(Tree tree) {
		// Compute the yield of the specified tree.
		List<String> yield = tree.getYield();

		// Normalize unknown words.
		List<String> tokens = new LinkedList<String>();
		for (String token : yield) {
			if (grammar.getCodeWord(token) < 0) {
				tokens.add(UNKNOWN);
			} else {
				tokens.add(token);
			}
		}

		// Parse the yield; measure the time used for parsing.
		long time = System.currentTimeMillis();
		Tree parse = parser.getBestParse(tokens);
		totalTime += System.currentTimeMillis() - time;
		nSentences++;

		// Call the fallback parser if need be.
		if (parse == null) {
			parse = fallbackParser.getBestParse(tokens);
		}

		// Put the original tokens back in place.
		parse = replaceYield(parse, yield);

		// Debinarize the parsed tree and update the scores.
		Tree postprocessed1 = TreeTransformers.getDebinarizer().transform(parse);
//		Tree postprocessed2 = TreeTransformers.getParentAnnotationStripper().transform(postprocessed1);
		bracketScorer.update(tree, postprocessed1);

		// Report progress.
		if (nSentences % 10 == 0) {
			reportProgress();
		}

		super.next(postprocessed1);
	}

	@Override
	public void exit() {
		super.exit();
		System.err.println();
		System.err.println("FINAL RESULTS:");
		reportProgress();
		System.err.println("Total time: " + (totalTime / 1000) + " seconds.");
	}

	private void reportProgress() {
		System.err.print("#sentences: ");
		System.err.print(nSentences);
		System.err.print(", F1: ");
		System.err.print(bracketScorer.getFMeasure());
		System.err.print(", seconds/sentence (avg.): ");
		System.err.println((totalTime / 1000) / nSentences);
	}

	/**
	 * Replaces the yield of the source tree with the specified yield.
	 *
	 * @param source the tree whose yield should be replaced
	 * @param yield the new yield
	 * @return the modified source tree
	 */
	private static Tree replaceYield(Tree source, List<String> yield) {
		Tree target = source.getCopy();
		List<Tree> terminalSubtrees = target.getLeaves();
		Iterator<String> yieldI = yield.iterator();
		for (Tree terminalSubtree : terminalSubtrees) {
			if (yieldI.hasNext()) {
				terminalSubtree.setLabel(yieldI.next());
			}
		}
		return target;
	}
}
