/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

import java.util.Deque;
import java.util.LinkedList;
import java.util.List;

/**
 * A phrase structure tree.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class Tree {

	/**
	 * The label of the root node of this tree.
	 */
	private String label;
	/**
	 * The immediate supertree of this tree.
	 */
	private Tree supertree;
	/**
	 * The left sibling tree of this tree.
	 */
	private Tree siblingL;
	/**
	 * The right sibling tree of this tree.
	 */
	private Tree siblingR;
	/**
	 * The number of immediate subtrees of this tree.
	 */
	private int nSubtrees;
	/**
	 * The leftmost immediate subtree of this tree.
	 */
	private Tree subtreeL;
	/**
	 * The rightmost immediate subtree of this tree.
	 */
	private Tree subtreeR;

	/**
	 * Constructs a new tree with the specified label
	 *
	 * @param label the label of the new tree
	 */
	public Tree(String label) {
		this.label = label;
	}

	/**
	 * Construct a new tree with a null label.
	 */
	public Tree() {
	}

	/**
	 * Returns the label of the root node of this tree.
	 *
	 * @return the label of the root node of this tree
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * Returns the immediate supertree of this tree.
	 *
	 * @return the immediate supertree of this tree
	 */
	public Tree getSupertree() {
		return supertree;
	}

	/**
	 * Tests whether this tree has any immediate supertrees.
	 *
	 * @return <code>true</code> iff this tree has no immediate supertrees
	 */
	public boolean isInitial() {
		return supertree == null;
	}

	/**
	 * Returns the left sibling tree of this tree.
	 *
	 * @return the left sibling tree of this tree, or <code>null</code> if this
	 * tree does not have any left sibling trees
	 */
	public Tree getLeftSibling() {
		return siblingL;
	}

	/**
	 * Tests whether this tree is at the leftmost periphery.
	 *
	 * @return <code>true</code> iff this tree is at the leftmost periphery
	 */
	public boolean isLeftmost() {
		return siblingL == null;
	}

	/**
	 * Returns the right sibling tree of this tree.
	 *
	 * @return the right sibling tree of this tree, or <code>null</code> if this
	 * tree does not have any right sibling trees
	 */
	public Tree getRightSibling() {
		return siblingR;
	}

	/**
	 * Tests whether this tree is at the rightmost periphery.
	 *
	 * @return <code>true</code> iff this tree is at the rightmost periphery
	 */
	public boolean isRightmost() {
		return siblingR == null;
	}

	/**
	 * Returns the number of immediate subtrees of this tree.
	 *
	 * @return the number of immediate subtrees of this tree
	 */
	public int getNSubtrees() {
		return nSubtrees;
	}

	/**
	 * Returns the leftmost immediate subtree of this tree.
	 *
	 * @return the leftmost immediate subtree of this tree
	 */
	public Tree getLeftmostSubtree() {
		return subtreeL;
	}

	/**
	 * Returns the rightmost immediate subtree of this tree.
	 *
	 * @return the rightmost immediate subtree of this tree
	 */
	public Tree getRightmostSubtree() {
		return subtreeR;
	}

	/**
	 * Tests whether this tree has any immediate subtrees.
	 *
	 * @return <code>true</code> iff this tree has no immediate subtrees
	 */
	public boolean isTerminal() {
		return nSubtrees == 0;
	}

	/**
	 * Returns the subtrees of this tree.
	 *
	 * @return the subtrees of this tree as a double-ended queue
	 */
	public Deque<Tree> getSubtrees() {
		Deque<Tree> subtrees = new LinkedList<Tree>();
		for (Tree tree = getLeftmostSubtree(); tree != null; tree = tree.getRightSibling()) {
			subtrees.addLast(tree);
		}
		return subtrees;
	}

	/**
	 * Replaces the label of the root node of this tree with the specified
	 * label.
	 *
	 * @param label the new label of the root node of this tree
	 */
	public void setLabel(String label) {
		this.label = label;
	}

	/**
	 * Makes the specified tree the only immediate subtree of this tree.
	 *
	 * @param tree the tree to be made the only immediate subtree of this tree
	 */
	private void makeSubtree(Tree tree) {
		assert nSubtrees == 0 && tree.supertree == null;
		this.nSubtrees = 1;
		this.subtreeL = tree;
		this.subtreeR = tree;
		tree.supertree = this;
	}

	/**
	 * Adds the specified tree as a left sibling to this tree.
	 *
	 * @param tree the tree to be a added as the left sibling of this tree
	 */
	public void addLeftSibling(Tree tree) {
		assert supertree != null && tree.supertree == null;
		supertree.nSubtrees++;
		if (siblingL == null) {
			// This tree is the leftmost subtree of its parent.
			supertree.subtreeL = tree;
		} else {
			siblingL.siblingR = tree;
		}
		tree.supertree = supertree;
		tree.siblingL = siblingL;
		tree.siblingR = this;
		this.siblingL = tree;
	}

	/**
	 * Adds the specified tree as a right sibling to this tree.
	 *
	 * @param tree the tree to be added as the right sibling of this tree
	 */
	public void addRightSibling(Tree tree) {
		assert supertree != null && tree.supertree == null;
		supertree.nSubtrees++;
		if (siblingR == null) {
			// This tree is the rightmost child of its parent.
			supertree.subtreeR = tree;
		} else {
			siblingR.siblingL = tree;
		}
		tree.supertree = supertree;
		tree.siblingL = this;
		tree.siblingR = siblingR;
		this.siblingR = tree;
	}

	/**
	 * Adds the specified tree as the leftmost subtree of this tree.
	 *
	 * @param tree the tree to be added as the leftmost subtree of this tree
	 */
	public void addAsLeftmostSubtree(Tree tree) {
		assert tree.supertree == null;
		if (nSubtrees == 0) {
			makeSubtree(tree);
		} else {
			subtreeL.addLeftSibling(tree);
		}
	}

	/**
	 * Adds the specified tree as the rightmost subtree of this tree.
	 *
	 * @param tree the tree to be added as the rightmost subtree of this tree
	 */
	public void addAsRightmostSubtree(Tree tree) {
		assert tree.supertree == null;
		if (nSubtrees == 0) {
			makeSubtree(tree);
		} else {
			subtreeR.addRightSibling(tree);
		}
	}

	/**
	 * Traverses this tree left-to-right, sending notifications to the specified
	 * handler.
	 *
	 * @param handler the handler to be notified during the traversal of this
	 * tree
	 */
	public void traverseLeftToRight(TraversalHandler handler) {
		traverse(new TraversalItemLR(this), handler);
	}

	/**
	 * Traverses this tree left-to-right, sending notifications to the specified
	 * handler. During the traversal, subtrees to whose initial tree the
	 * specified predicate applies are ignored.
	 *
	 * @param predicate specifies subtrees that should be ignored during the
	 * traversal
	 * @param handler the handler to be notified during the traversal of this
	 * tree
	 */
	public void traverseLeftToRight(TreePredicate predicate, TraversalHandler handler) {
		traverse(new IgnoreSubtreeItem(predicate, new TraversalItemLR(this)), handler);
	}

	/**
	 * Traverses this tree right-to-left, sending notifications to the specified
	 * handler.
	 *
	 * @param handler the handler to be notified during the traversal of this
	 * tree
	 */
	public void traverseRightToLeft(TraversalHandler handler) {
		traverse(new TraversalItemRL(this), handler);
	}

	/**
	 * Traverses this tree right-to-left, sending notifications to the specified
	 * handler. During the traversal, subtrees to whose initial tree the
	 * specified predicate applies are ignored.
	 *
	 * @param predicate specifies subtrees that should be ignored during the
	 * traversal
	 * @param handler the handler to be notified during the traversal of this
	 * tree
	 */
	public void traverseRightToLeft(TreePredicate predicate, TraversalHandler handler) {
		traverse(new IgnoreSubtreeItem(predicate, new TraversalItemRL(this)), handler);
	}

	/**
	 * Traverses the tree defined by the specified traversal item, sending
	 * notifications to the specified handler.
	 *
	 * @param handler the handler to be notified during the traversal of this
	 * tree
	 * @param root the root item
	 */
	private static void traverse(TraversalItem root, TraversalHandler handler) {
		Deque<TraversalItem> agenda = new LinkedList<TraversalItem>();
		agenda.addLast(root);
		handler.enter(root.getTree());
		while (!agenda.isEmpty()) {
			TraversalItem item = agenda.removeLast();
			if (item.hasNext()) {
				TraversalItem next = item.getNext();
				agenda.addLast(item);
				handler.enter(next.getTree());
				agenda.addLast(next);
			} else {
				handler.leave(item.getTree());
			}
		}
	}

	/**
	 * Returns a copy of this tree.
	 *
	 * @return a copy of this tree
	 */
	public Tree getCopy() {
		CopyHandler handler = new CopyHandler();
		traverseLeftToRight(handler);
		return handler.getTree();
	}

	/**
	 * Returns a copy of this tree, ignoring subtrees to which the specified
	 * predicate applies.
	 *
	 * @param predicate specifies subtrees that should not be copied
	 * @return a copy of this tree
	 */
	public Tree getCopy(TreePredicate predicate) {
		CopyHandler handler = new CopyHandler();
		traverseLeftToRight(predicate, handler);
		return handler.getTree();
	}

	/**
	 * Returns a copy of this tree, with all nodes to which the specified
	 * predicate applies spliced out.
	 *
	 * @param predicate specifies subtrees whose root nodes should be spliced
	 * out
	 * @return a copy of this tree
	 */
	public Tree splice(TreePredicate predicate) {
		assert !predicate.applies(this);
		CopyHandler copyHandler = new CopyHandler();
		IgnoreHandler ignoreHandler = new IgnoreHandler(predicate, copyHandler);
		traverseLeftToRight(ignoreHandler);
		return copyHandler.getTree();
	}

	/**
	 * Returns the yield of this tree.
	 *
	 * @return the yield of this tree
	 */
	public List<String> getYield() {
		YieldHandler handler = new YieldHandler();
		traverseLeftToRight(handler);
		return handler.getYield();
	}

	/**
	 * Returns the leaves of this tree.
	 *
	 * @return the leaves of this tree
	 */
	public List<Tree> getLeaves() {
		GetLeavesHandler handler = new GetLeavesHandler();
		traverseLeftToRight(handler);
		return handler.getLeaves();
	}

	/**
	 * Returns a string representation of this tree.
	 *
	 * @return a string representation of this tree
	 */
	@Override
	public String toString() {
		ToStringHandler handler = new ToStringHandler();
		traverseLeftToRight(handler);
		return handler.getString();
	}

	private interface TraversalItem {

		/**
		 * Returns the tree associated with this item.
		 *
		 * @return the tree associated with this item
		 */
		public abstract Tree getTree();

		/**
		 * Tests whether this item can spawn off a new item.
		 *
		 * @return <code>true</code> iff this item can spawn off a new item
		 */
		public abstract boolean hasNext();

		/**
		 * Returns the next item spawned off by this item.
		 *
		 * @return the next item spawned off by this item
		 */
		public abstract TraversalItem getNext();
	}

	/**
	 * Item for left-to-right traversal.
	 */
	private class TraversalItemLR implements TraversalItem {

		private final Tree tree;
		private Tree child;

		public TraversalItemLR(Tree node) {
			this.tree = node;
			this.child = node.subtreeL;
		}

		public Tree getTree() {
			return tree;
		}

		public boolean hasNext() {
			return child != null;
		}

		public TraversalItem getNext() {
			TraversalItemLR next = new TraversalItemLR(child);
			this.child = child.getRightSibling();
			return next;
		}
	}

	/**
	 * Item for right-to-left traversal.
	 */
	private class TraversalItemRL implements TraversalItem {

		private final Tree tree;
		private Tree child;

		public TraversalItemRL(Tree node) {
			this.tree = node;
			this.child = node.subtreeR;
		}

		public Tree getTree() {
			return tree;
		}

		public boolean hasNext() {
			return child != null;
		}

		public TraversalItemRL getNext() {
			TraversalItemRL next = new TraversalItemRL(child);
			this.child = child.getLeftSibling();
			return next;
		}
	}

	/**
	 * Item that ignores subtrees matched by a predicate.
	 */
	private class IgnoreSubtreeItem implements TraversalItem {

		private final TraversalItem delegate;
		private final TreePredicate ignore;
		private TraversalItem next;

		public IgnoreSubtreeItem(TreePredicate ignore, TraversalItem delegate) {
			this.ignore = ignore;
			this.delegate = delegate;
			updateNext();
		}

		public Tree getTree() {
			return delegate.getTree();
		}

		public boolean hasNext() {
			return next != null;
		}

		public TraversalItem getNext() {
			TraversalItem result = next;
			updateNext();
			return result;
		}

		private void updateNext() {
			this.next = null;
			while (delegate.hasNext()) {
				TraversalItem delegateNext = delegate.getNext();
				if (!ignore.applies(delegateNext.getTree())) {
					this.next = new IgnoreSubtreeItem(ignore, delegateNext);
					break;
				}
			}
		}
	}

	/**
	 * Handler that copies the tree that it receives notications about.
	 */
	private class CopyHandler implements TraversalHandler {

		private final Deque<Tree> agenda;
		private Tree last;

		public CopyHandler() {
			this.agenda = new LinkedList<Tree>();
		}

		public Tree getTree() {
			assert agenda.isEmpty();
			return last;
		}

		public void enter(Tree tree) {
			last = new Tree(tree.getLabel());
			agenda.addLast(last);
		}

		public void leave(Tree tree) {
			last = agenda.removeLast();
			if (!agenda.isEmpty()) {
				agenda.getLast().addAsRightmostSubtree(last);
			}
		}
	}

	/**
	 * Handler that ignores notifications about trees matching a predicate.
	 */
	private class IgnoreHandler implements TraversalHandler {

		private final TreePredicate predicate;
		private final TraversalHandler delegate;

		public IgnoreHandler(TreePredicate predicate, TraversalHandler handler) {
			this.predicate = predicate;
			this.delegate = handler;
		}

		public void enter(Tree tree) {
			if (!predicate.applies(tree)) {
				delegate.enter(tree);
			}
		}

		public void leave(Tree tree) {
			if (!predicate.applies(tree)) {
				delegate.leave(tree);
			}
		}
	}

	/**
	 * Handler used to collect the yield of a tree.
	 */
	private class YieldHandler implements TraversalHandler {

		private final List<String> tokens;

		public YieldHandler() {
			this.tokens = new LinkedList<String>();
		}

		public List<String> getYield() {
			return tokens;
		}

		public void enter(Tree tree) {
			if (tree.isTerminal()) {
				tokens.add(tree.getLabel());
			}
		}

		public void leave(Tree tree) {
		}
	}

	/**
	 * Handler used to collect the leaves off a tree.
	 */
	private class GetLeavesHandler implements TraversalHandler {

		private final List<Tree> leaves;

		public GetLeavesHandler() {
			this.leaves = new LinkedList<Tree>();
		}

		public List<Tree> getLeaves() {
			return leaves;
		}

		public void enter(Tree tree) {
			if (tree.isTerminal()) {
				leaves.add(tree);
			}
		}

		public void leave(Tree tree) {
		}
	}

	/**
	 * Handler used to compute the string representation of a tree.
	 */
	private class ToStringHandler implements TraversalHandler {

		private final StringBuilder builder;

		public ToStringHandler() {
			this.builder = new StringBuilder();
		}

		public String getString() {
			return builder.toString();
		}

		public void enter(Tree tree) {
			if (!tree.isInitial()) {
				builder.append(" ");
			}
			if (tree.isTerminal()) {
				builder.append(tree.getLabel());
			} else {
				builder.append("(");
				builder.append(tree.getLabel());
			}
		}

		public void leave(Tree tree) {
			if (!tree.isTerminal()) {
				builder.append(")");
			}
		}
	}
}
