/*
 * See the file "LICENSE" for the full license governing this code.
 */
package up;

import java.util.Arrays;
import java.util.Deque;
import java.util.LinkedList;

/**
 * A collection of tree transformers.
 *
 * @author Marco Kuhlmann <marco.kuhlmann@lingfil.uu.se>
 */
public class TreeTransformers {

	/**
	 * The prefix of labels introduced during binarization.
	 */
	private static final String BIN_LABEL_PREFIX = "@BIN_";
	/**
	 * The delimiter of labels assembled during binarization.
	 */
	private static final String BIN_LABEL_DELIMITER = "_";
	/**
	 * Shared instance of the TerminalNodesStripper.
	 */
	private static TreeTransformer terminalNodesStripper;
	/**
	 * Shared instance of the FunctionAnnotationStripper.
	 */
	private static TreeTransformer functionAnnotationStripper;
	/**
	 * Shared instance of the EmptySubtreeStripper.
	 */
	private static TreeTransformer emptySubtreeStripper;
	/**
	 * Shared instance of the UnaryChainsSplicer.
	 */
	private static TreeTransformer unaryChainsSplicer;
	/**
	 * Shared instance of the Binarizer.
	 */
	private static TreeTransformer binarizer;
	/**
	 * Shared instance of the Debinarizer.
	 */
	private static TreeTransformer debinarizer;
	/**
	 * Shared instance of the ParentAnnotator.
	 */
	private static TreeTransformer parentAnnotator;
	/**
	 * Shared instance of the ParentAnnotationStripper.
	 */
	private static TreeTransformer parentAnnotationStripper;

	/**
	 * Private constructor to prevent instantiation.
	 */
	private TreeTransformers() {
	}

	/**
	 * Returns the shared instance of the
	 * <code>FunctionAnnotationStripper</code>.
	 *
	 * @return the shared instance of the
	 * <code>FunctionAnnotationStripper</code>
	 */
	public static TreeTransformer getFunctionAnnotationStripper() {
		if (functionAnnotationStripper == null) {
			functionAnnotationStripper = new FunctionAnnotationStripper();
		}
		return functionAnnotationStripper;
	}

	/**
	 * Returns the shared instance of the
	 * <code>EmptySubtreeStripper</code>.
	 *
	 * @return the shared instance of the
	 * <code>EmptySubtreeStripper</code>
	 */
	public static TreeTransformer getEmptySubtreeStripper() {
		if (emptySubtreeStripper == null) {
			emptySubtreeStripper = new EmptySubtreeStripper();
		}
		return emptySubtreeStripper;
	}

	/**
	 * Returns the shared instance of the
	 * <code>UnaryChainsSplicer</code>.
	 *
	 * @return the shared instance of the
	 * <code>UnaryChainsSplicer</code>
	 */
	public static TreeTransformer getUnaryChainsSplicer() {
		if (unaryChainsSplicer == null) {
			unaryChainsSplicer = new UnaryChainsSplicer();
		}
		return unaryChainsSplicer;
	}

	/**
	 * Returns the shared instance of the
	 * <code>Binarizer</code>.
	 *
	 * @return the shared instance of the
	 * <code>Binarizer</code>
	 */
	public static TreeTransformer getBinarizer() {
		if (binarizer == null) {
			binarizer = new Binarizer();
		}
		return binarizer;
	}

	/**
	 * Returns the shared instance of the
	 * <code>Debinarizer</code>.
	 *
	 * @return the shared instance of the
	 * <code>Debinarizer</code>
	 */
	public static TreeTransformer getDebinarizer() {
		if (debinarizer == null) {
			debinarizer = new Debinarizer();
		}
		return debinarizer;
	}

	/**
	 * Returns the shared instance of the
	 * <code>ParentAnnotator</code>.
	 *
	 * @return the shared instance of the
	 * <code>ParentAnnotator</code>
	 */
	public static TreeTransformer getParentAnnotator() {
		if (parentAnnotator == null) {
			parentAnnotator = new ParentAnnotator();
		}
		return parentAnnotator;
	}

	/**
	 * Returns the shared instance of the
	 * <code>ParentAnnotationStripper</code>.
	 *
	 * @return the shared instance of the
	 * <code>ParentAnnotationStripper</code>
	 */
	public static TreeTransformer getParentAnnotationStripper() {
		if (parentAnnotationStripper == null) {
			parentAnnotationStripper = new ParentAnnotationStripper();
		}
		return parentAnnotationStripper;
	}

	/**
	 * A transformer that removes the function annotation from PTB labels.
	 */
	private static class FunctionAnnotationStripper implements TreeTransformer {

		private static final TraversalHandler HANDLER = new Handler();

		/**
		 * Removes the function annotation from the labels of the specified
		 * tree.
		 *
		 * @param tree the tree to be transformed
		 * @return a copy of the specified tree in which all function annotation
		 * has been deleted
		 */
		public Tree transform(Tree tree) {
			Tree copy = tree.getCopy();
			copy.traverseLeftToRight(HANDLER);
			return copy;
		}

		private static class Handler implements TraversalHandler {

			public Handler() {
			}

			public void enter(Tree tree) {
				tree.setLabel(getStrippedLabel(tree.getLabel()));
			}

			public void leave(Tree tree) {
			}

			private static String getStrippedLabel(String label) {
				int[] positions = new int[3];
				if (label.startsWith("-")) {
					return label;
				} else {
					positions[0] = label.indexOf('-');
					positions[1] = label.indexOf('=');
					positions[2] = label.indexOf('^');
					Arrays.sort(positions);
					for (int i = 0; i < positions.length; i++) {
						if (positions[i] >= 0) {
							return label.substring(0, positions[i]);
						}
					}
					return label;
				}
			}
		}
	}

	/**
	 * A transformer that removes subtrees that dominate the empty string.
	 */
	private static class EmptySubtreeStripper implements TreeTransformer {

		/**
		 * Removes empty subtrees from the specified PTB tree. A subtree is
		 * called empty if it dominates the empty string.
		 *
		 * @param tree the tree to be transformed
		 * @return a copy of the specified tree in which all empty subtrees have
		 * been deleted
		 */
		public Tree transform(Tree tree) {
			return tree.getCopy(new TreePredicate() {

				public boolean applies(Tree tree) {
					return tree.getLabel().equals("-NONE-") || (tree.getNSubtrees() == 1 && applies(tree.getLeftmostSubtree()));
				}
			});
		}
	}

	/**
	 * A transformer that removes unary chains.
	 */
	private static class UnaryChainsSplicer implements TreeTransformer {

		/**
		 * Removes unary chains from the specified PTB tree.
		 *
		 * @param tree the tree to be transformed
		 * @return a copy of the specified tree in which all unary chains have
		 * been deleted
		 */
		public Tree transform(Tree tree) {
			return tree.splice(new TreePredicate() {

				public boolean applies(Tree tree) {
					return tree.getNSubtrees() == 1 && !tree.getLeftmostSubtree().isTerminal() && tree.getLabel().equals(tree.getLeftmostSubtree().getLabel());
				}
			});
		}
	}

	/**
	 * A transformer that binarizes a given tree. The specific binarization
	 * performed is a lossless, left-branching binarization.
	 */
	private static class Binarizer implements TreeTransformer {

		/**
		 * Binarizes the specified tree.
		 *
		 * @param tree the tree to be transformed
		 * @return a binarized copy of the tree
		 */
		public Tree transform(Tree tree) {
			BinarizeLHandler handler = new BinarizeLHandler();
			tree.traverseRightToLeft(handler);
			return handler.getTree();
		}

		private static class BinarizeLHandler implements TraversalHandler {

			private final Deque<Tree> agenda;
			private Tree last;

			public BinarizeLHandler() {
				this.agenda = new LinkedList<Tree>();
			}

			public Tree getTree() {
				assert agenda.isEmpty();
				return last;
			}

			public void enter(Tree tree) {
				if (!tree.isInitial() && !tree.isLeftmost() && !tree.isRightmost()) {
					String label = tree.getLabel();
					agenda.addLast(new Tree());
				}
				agenda.addLast(new Tree(tree.getLabel()));
			}

			public void leave(Tree tree) {
				if (tree.getNSubtrees() > 2) {
					String label = agenda.getLast().getLeftmostSubtree().getLabel();

					for (int i = 0; i < tree.getNSubtrees() - 2; i++) {
						Tree hook = agenda.removeLast();

						StringBuilder b1 = new StringBuilder();
						b1.append(label);
						b1.append(BIN_LABEL_DELIMITER);
						b1.append(hook.getRightmostSubtree().getLabel());
						label = b1.toString();

						String taggedLabel = getTaggedLabel(tree, label);
						hook.setLabel(taggedLabel);

						agenda.getLast().addAsLeftmostSubtree(hook);
					}
				}

				last = agenda.removeLast();
				if (!agenda.isEmpty()) {
					agenda.getLast().addAsLeftmostSubtree(last);
				}
			}

			private static String getTaggedLabel(Tree tree, String label) {
				StringBuilder b = new StringBuilder();
				b.append(BIN_LABEL_PREFIX);
				b.append(tree.getLabel());
				b.append(BIN_LABEL_DELIMITER);
				b.append(label);
				return b.toString();
			}
		}
	}

	/**
	 * A transformer that debinarizes a given tree.
	 */
	private static class Debinarizer implements TreeTransformer {

		/**
		 * Debinarizes the specified tree.
		 *
		 * @param tree the tree to be transformed
		 * @return a debinarized copy of the tree
		 */
		public Tree transform(Tree tree) {
			return tree.splice(new TreePredicate() {

				public boolean applies(Tree tree) {
					return tree.getLabel().startsWith(BIN_LABEL_PREFIX);
				}
			});
		}
	}

	/**
	 * A transformer that adds parent annotations.
	 */
	private static class ParentAnnotator implements TreeTransformer {

		private static final TraversalHandler HANDLER = new Handler();

		/**
		 * Adds parent annotations to the specified tree. tree.
		 *
		 * @param tree the tree to be transformed
		 * @return a copy of the specified tree in which all nodes except for
		 * the root node and leaf nodes have been parent-annotated
		 */
		public Tree transform(Tree tree) {
			Tree copy = tree.getCopy();
			copy.traverseLeftToRight(HANDLER);
			return copy;
		}

		private static class Handler implements TraversalHandler {

			public Handler() {
			}

			public void enter(Tree tree) {
			}

			public void leave(Tree tree) {
				if (!tree.isTerminal() && !tree.isInitial() && !(tree.getNSubtrees() == 1 && tree.getLeftmostSubtree().isTerminal())) {
					String parentLabel = tree.getSupertree().getLabel();
					tree.setLabel(tree.getLabel() + "^" + parentLabel);
				}
			}
		}
	}

	/**
	 * A transformer that removes the parent annotation.
	 */
	private static class ParentAnnotationStripper implements TreeTransformer {

		private static final TraversalHandler HANDLER = new Handler();

		/**
		 * Removes the parent annotation from the labels of the specified tree.
		 *
		 * @param tree the tree to be transformed
		 * @return a copy of the specified tree in which all parent annotation
		 * has been deleted
		 */
		public Tree transform(Tree tree) {
			Tree copy = tree.getCopy();
			copy.traverseLeftToRight(HANDLER);
			return copy;
		}

		private static class Handler implements TraversalHandler {

			public Handler() {
			}

			public void enter(Tree tree) {
				int p = tree.getLabel().indexOf("^");
				if (p >= 0) {
					tree.setLabel(tree.getLabel().substring(0, p));
				}
			}

			public void leave(Tree tree) {
			}
		}
	}
}
